/*
 * powerthend.c: (c) 2005-2006 Peter Schloemer <powerthend@scheissname.de>
 *
 * Released under the GNU General Public License. See the LICENSE file
 * included with this file.
 *
 * Based on powernowd (c) 2003-2006 John Clemens <clemej@alum.rpi.edu>
 * 	As I don't know which contributions this may include, here is a full
 * 	list as given by John Clemens:
 *
 * 	Warren Togami <warren@togami.com>
 * 	Michael Schreiter <Michael_Schreiter@gmx.de>
 * 	Wolfgang Tremmel <tremmel@garf.de>
 * 	Hans Ulrich Niedermann <kdev@n-dimensional.de>
 * 	Bdale Garbee, Vince Weaver
 * 	-- And many others who have sent me patches over the past 3 years,
 * 	   some of whom I've never even acknowledged.  Just know that your
 * 	   support is appreciated, even if I didn't merge in your ideas.
 * 	   Thank you.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>
#include <time.h>

#define pprintf(level, ...) do { \
	if (level <= verbosity) { \
		if (daemonize) \
			syslog(LOG_INFO, __VA_ARGS__); \
		else \
			printf(__VA_ARGS__); \
	} \
} while(0)

typedef struct cpustats {
	unsigned long long user;
	unsigned long long mynice;
	unsigned long long system;
	unsigned long long idle;
	unsigned long long iowait;
	unsigned long long irq;
	unsigned long long softirq;
} cpustats_t;

typedef struct cpuinfo {
	unsigned int cpuid;
	unsigned int max_speed;
	unsigned int min_speed;
	unsigned int current_speed;
	int fd;
	char *sysfs_dir;
	cpustats_t *last_reading;
	cpustats_t *reading;
        unsigned long *freq_table;
        int table_size;
	int threads_per_core;
	int scalable_unit;
} cpuinfo_t;

/*
 * Global array that contains a pointer to all CPU info.
 */

cpuinfo_t **all_cpus;

/* idea stolen from procps */
static char buf[1024];

/* for a daemon as simple as this, global data is ok. */
/* settings */
int daemonize = 1;
int ignore_nice = 1;
int verbosity = 0;
unsigned int poll = 100; /* in msecs */
unsigned int highwater = 95;
unsigned int max_limit = 0;
unsigned int min_limit = 0;
unsigned int t_per_core = 1;
unsigned int cores_specified = 0;

#define SYSFS_TREE "/sys/devices/system/cpu/"
#define SYSFS_SETSPEED "scaling_setspeed"

#define VERSION	"1.1.1"

void help(void)
{

	printf("PowerThen Daemon v%s, (c) 2005-2006 Peter Schloemer\n", VERSION);
	printf("Based on PowerNowd (c) 2003-2006 John Clemens\n");
	printf("Daemon to control the speed and voltage of cpus.\n\n");
	printf("This is a simple client to the CPUFreq driver, and uses\n");
	printf("linux kernel v2.5 sysfs interface.  You need a supported\n");
	printf("cpu, and a kernel that supports sysfs to run this daemon.\n");
	printf("\nAvailable Options:\n");
	printf(" 	-h	Print this help message\n");
	printf("	-d	Don't detach from terminal (default is to\n");
	printf("		detach and run in the background)\n");
	printf("	-v	Increase output verbosity, can be used more than once.\n");
	printf("	-q	Quiet mode, only emergency output.\n");
	printf("	-n	Include 'nice'd processes in calculations\n");
	printf("	-p #	Polling frequency in msecs (default = 100)\n");
	printf("	-c #	Specify number of threads per power-managed core\n");
	printf("	-u #	CPU usage upper limit percentage [0 .. 100, default 95]\n");

	printf("\n");
	return;
}

/*
 * Open a file and copy it's first 1024 bytes into the global "buf".
 * Zero terminate the buffer.
 */
int read_file(const char *file, int fd, int new)
{
	int n, err;

	if (new) {
		if ((fd = open(file, O_RDONLY)) == -1) {
			err = errno;
			perror(file);
			return err;
		}
	}

	lseek(fd, 0, SEEK_SET);
	if ((n = read(fd, buf, sizeof(buf)-1)) < 0) {
		err = errno;
		perror(file);
		close(fd);
		return err;
	}
	buf[n] = '\0';

	if (new)
		close(fd);

	return 0;
}

/*
 * Reads /proc/stat into buf, and parses the output.
 *
 * Format of line:
 * ...
 * cpu<id> <user> <nice> <system> <idle> <iowait> <irq> <softirq>
 */
int get_stat(cpuinfo_t *cpu)
{
	char *p1, *p2, searchfor[10];
	int err;

	if ((err = read_file("/proc/stat", cpu->fd, 0)) != 0) {
		return err;
	}

	sprintf(searchfor, "cpu%d ", cpu->cpuid);

	p1 = strstr(buf, searchfor);
	if (p1 == NULL) {
		perror("Error parsing /proc/stat");
		return ENOENT;
	}

	p2 = p1+strlen(searchfor);

	memcpy(cpu->last_reading, cpu->reading, sizeof(cpustats_t));

	cpu->reading->user = strtol(p2, &p2, 10);
	cpu->reading->mynice = strtol(p2, &p2, 10);
	cpu->reading->system = strtol(p2, &p2, 10);
	cpu->reading->idle = strtol(p2, &p2, 10);
	cpu->reading->iowait = strtol(p2, &p2, 10);
	cpu->reading->irq = strtol(p2, &p2, 10);
	cpu->reading->softirq = strtol(p2, &p2, 10);

	return 0;
}

/*
 * Updates cpu_info (current frequency).
 */
int update_cpu_info(cpuinfo_t *cpu)
{
	char scratch[100];
	int err;

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, "scaling_cur_freq", 18);
	if ((err = read_file(scratch, 0, 1)) != 0) {
		return err;
	}

	cpu->current_speed = strtol(buf, NULL, 10);

	return 0;
}

/*
 * Once a decision is made, change the speed.
 */

int change_speed(cpuinfo_t *cpu, unsigned int speed)
{
	int fd, len, err;
	char writestr[100];

	if (cpu->cpuid != cpu->scalable_unit)
		return 0;

	strncpy(writestr, cpu->sysfs_dir, 50);
	strncat(writestr, SYSFS_SETSPEED, 20);

	if ((fd = open(writestr, O_WRONLY)) < 0) {
		err = errno;
		perror("Can't open scaling_setspeed");
		return err;
	}

	lseek(fd, 0, SEEK_CUR);

	sprintf(writestr, "%d\n", speed);

	pprintf(4,"speed=%d, str=%s", speed, writestr);

	if ((len = write(fd, writestr, strlen(writestr))) < 0) {
		err = errno;
		perror("Couldn't write to scaling_setspeed\n");
		close(fd);
		return err;
	}

	if (len != strlen(writestr)) {
		printf("Could not write scaling_setspeed\n");
		close(fd);
		return EPIPE;
	}
	close(fd);

	update_cpu_info(cpu);

	return 0;
}

/*
 * The heart of the program... decide to raise or lower the speed.
 */
unsigned int inline decide_speed(cpuinfo_t *cpu)
{
	int err;
	float pct;
	unsigned long long usage, total;

	if ((err = get_stat(cpu)) < 0) {
		perror("Can't get stats");
		return err;
	}

	total = (cpu->reading->user - cpu->last_reading->user) +
		(cpu->reading->system - cpu->last_reading->system) +
		(cpu->reading->mynice - cpu->last_reading->mynice) +
		(cpu->reading->idle - cpu->last_reading->idle) +
		(cpu->reading->iowait - cpu->last_reading->iowait) +
		(cpu->reading->irq - cpu->last_reading->irq) +
		(cpu->reading->softirq - cpu->last_reading->softirq);

	if (ignore_nice) {
		usage = (cpu->reading->user - cpu->last_reading->user) +
			(cpu->reading->system - cpu->last_reading->system) +
			(cpu->reading->irq - cpu->last_reading->irq) +
			(cpu->reading->softirq - cpu->last_reading->softirq);
	} else {
		usage = (cpu->reading->user - cpu->last_reading->user) +
			(cpu->reading->mynice - cpu->last_reading->mynice) +
			(cpu->reading->system - cpu->last_reading->system) +
			(cpu->reading->irq - cpu->last_reading->irq) +
			(cpu->reading->softirq - cpu->last_reading->softirq);
	}

	pct = ((float)usage)/((float)total);

	pprintf(4,"PCT = %f\n", pct);

	/* update cpu frequence */
	update_cpu_info(cpu);

	return cpu->current_speed * (pct) / ((float)highwater/100);
}

/*
 * Abuse glibc's qsort.  Compare function to sort list of frequencies in
 * ascending order.
 */
int faked_compare(const void *a, const void *b)
{
	unsigned long *a1 = (unsigned long *)a;
	unsigned long *b1 = (unsigned long *)b;

	if (*a1 < *b1) return 1;
	if (*a1 > *b1) return -1;

	return 0;
}

/*
 * Allocates and initialises the per-cpu data structures.
 */
int get_per_cpu_info(cpuinfo_t *cpu, int cpuid)
{
	char cpustr[100], scratch[100], tmp[10];
	int fd, err;

	cpu->cpuid = cpuid;
	cpu->sysfs_dir = (char *)malloc(50*sizeof(char));
	if (cpu->sysfs_dir == NULL) {
		perror("Couldn't allocate per-cpu sysfs_dir");
		return ENOMEM;
	}
	memset(cpu->sysfs_dir, 0, (50*sizeof(char)));

	strncpy(cpu->sysfs_dir, SYSFS_TREE, 30);
	sprintf(cpustr, "cpu%d/cpufreq/", cpuid);
	strncat(cpu->sysfs_dir, cpustr, 20);

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, "cpuinfo_max_freq", 18);
	if ((err = read_file(scratch, 0, 1)) != 0) {
		return err;
	}

	cpu->max_speed = strtol(buf, NULL, 10);

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, "cpuinfo_min_freq", 18);

	if ((err = read_file(scratch, 0, 1)) != 0) {
		return err;
	}

	cpu->min_speed = strtol(buf, NULL, 10);

	/* XXXjc read the real current speed */
	cpu->current_speed = cpu->max_speed;

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, "scaling_available_frequencies", 50);

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, "scaling_governor", 20);

	if ((err = read_file(scratch, 0, 1)) != 0) {
		perror("couldn't open scaling_governors file");
		return err;
	}

	if (strncmp(buf, "userspace", 9) != 0) {
		if ((fd = open(scratch, O_RDWR)) < 0) {
			err = errno;
			perror("couldn't open govn's file for writing");
			return err;
		}
		strncpy(tmp, "userspace\n", 11);
		if (write(fd, tmp, 11*sizeof(char)) < 0) {
			err = errno;
			perror("Error writing file governor");
			close(fd);
			return err;
		}
		if ((err = read_file(scratch, fd, 0)) != 0) {
			perror("Error reading back governor file");
			close(fd);
			return err;
		}
		close(fd);
		if (strncmp(buf, "userspace", 9) != 0) {
			perror("Can't set to userspace governor, exiting");
			return EPIPE;
		}
	}

	cpu->last_reading = (cpustats_t *)malloc(sizeof(cpustats_t));
	cpu->reading = (cpustats_t *)malloc(sizeof(cpustats_t));
	memset(cpu->last_reading, 0, sizeof(cpustats_t));
	memset(cpu->reading, 0, sizeof(cpustats_t));

	if ((cpu->fd = open("/proc/stat", O_RDONLY)) < 0) {
		err = errno;
		perror("can't open /proc/stat");
		return err;
	}

	if ((err = get_stat(cpu)) < 0) {
		perror("can't read /proc/stat");
		return err;
	}

	return 0;
}

/*
 * Signal handler for SIGTERM/SIGINT... clean up after ourselves
 */
void terminate(int signum)
{
	int ncpus, i;
	cpuinfo_t *cpu;

	ncpus = sysconf(_SC_NPROCESSORS_CONF);
	if (ncpus < 1) ncpus = 1;

	/*
	 * for each cpu, force it back to full speed.
	 * don't mix this with the below statement.
	 *
	 * 5 minutes ago I convinced myself you couldn't
	 * mix these two, now I can't remember why.
	 */
	for(i = 0; i < ncpus; i++) {
		cpu = all_cpus[i];
		change_speed(cpu, cpu->max_speed);
	}

	for(i = 0; i < ncpus; i++) {
		cpu = all_cpus[i];
		/* close the /proc/stat fd */
		close(cpu->fd);
		/* deallocate everything */
		free(cpu->sysfs_dir);
		free(cpu->last_reading);
		free(cpu->reading);
		free(cpu);
	}
	pprintf(0,"PowerThen Daemon Exiting.\n");

	closelog();

	exit(0);
}

/* Generic x86 cpuid function lifted from kernel sources */
/*
 * Generic CPUID function
 */
static inline void cpuid(int op, int *eax, int *ebx, int *ecx, int *edx)
{
        __asm__("cpuid"
                : "=a" (*eax),
                  "=b" (*ebx),
                  "=c" (*ecx),
                  "=d" (*edx)
                : "0" (op));
}

/*
 * A little bit of black magic to try and detect the number of cores per
 * processor.  This will have to be added on to for every architecture,
 * as we learn how to detect them from userspace.  Note: This method
 * assumes uniform processor/thread ID's.  First look for affected_cpus,
 * if not, then fall back to the cpuid check, or just default to 1.
 *
 * By 'thread' in this case, I mean 'entity that is part of one scalable
 * instance'.  For example, a P4 with hyperthreading has 2 threads in
 * one scalable instace.  So does an Athlon X2 dual core, because each
 * core has to have the same speed.  The new Yonah, on the other hand, may
 * have two scalable elements, as rumors say you can control both cores'
 * speed individually.  Lets hope the speedstep driver populates affected_cpus
 * correctly...
 *
 * You can always override this by using the -c command line option to
 * specify the number of threads per core.  If you do so, it will do a static
 * mapping, uniform for all real processors in the system.  Actually, so
 * will this one, because there's no way for me to bind to a processor.
 * (yet. :)
 */
int determine_threads_per_core(int ncpus)
{
	char filename[100], *p1;
	int err, count;

	/* if ncpus is one, we don't care */
	if (ncpus == 1) return 1;

	/*
	 * First look for the affected_cpus file, and count the
	 * number of cpus that supports.  Assume this is true for all
	 * cpus on the system.
	 */
	strncpy(filename, SYSFS_TREE, 30);
	sprintf(filename, "%scpu0/cpufreq/affected_cpus", filename);

	/*
	 * OK, the funkiest system I can think of right now is
	 * Sun's Niagara processor, which I think would have 32
	 * "cpus" in one scalable element.  So make this robust
	 * enough to at least handle more than 32 affected cpus
	 * at once.
	 *
	 * NOTE: I don't even know if Niagara supports scaling,
	 * I'm dealing with hypotheticals.
	 */

	count = 1;

	if ((err = read_file(filename, 0, 1)) == 0) {
		p1 = buf;
		err = strtoul(p1, &p1, 5);
		/*
		 * The first cpu should always be 0, so err should be 0
		 * after the first read.  If its anything else, default to
		 * one, print a message, and move on.
		 */
		if (err != 0) {
			pprintf(0, "WARN: cpu0 scaling doesn't affect cpu 0?"
				       " Assuming 1 thread per core.\n");
			return 1;
		}
		while ((err = strtol(p1, &p1, 5)) != 0)
			count ++;
		pprintf(1, "about to return count = %d\n", count);
		return count;
	}
	pprintf(0,"err=%d", err);

#ifdef __i386__
	/* Only get here if there's no affected_cpus file. */
	/*
	 * XXXjc fix eventually to run on each processor so you
	 * can support mixed multi and single-core cpus. Need to know
	 * how to force ourselves to run on one particular processor.
	 */
	int eax,ebx,ecx,edx, num=1;

	cpuid(1,&eax,&ebx,&ecx,&edx);

	/*
	 * Do we support hyperthreading?
	 * AMD's dual-core will masquerade as HT, so this should work
	 * for them too. (update: it does but this doesn't emulate
	 * the extra ebx parameter, appearently).
	 */
	if(edx & 0x08000000) {
		/*
		 * if so, is it enabled? If so, how many threads
		 * are enabled?  Thank you, sandpile.org and the LKML.
		 */
		num = (ebx & 0x00FF0000) >> 16;
	}
	/*
	 * if num = 0, default to 1.  Other non-multiples of cpus will
	 * be taken care of later
	 * XXXjc, rewrite to use sysfs (affected_cpus).
	 */
	return ((num)?num:1);
#endif
	/* always default to one thread per core */
	return 1;
}


/*
 * Main program loop.. parse arguments, sanity chacks, setup signal handlers
 * and then enter main loop
 */
int main(int argc, char **argv)
{
	cpuinfo_t *cpu;
	int ncpus, i, j, err, num_real_cpus, threads_per_core, cpubase;
	unsigned int change, change2;

	/* Parse command line args */
	while(1) {
		int c;

		c = getopt(argc, argv, "dnvqm:s:p:c:u:l:U:L:h");
		if (c == -1)
			break;

		switch(c) {
			case 'd':
				daemonize = 0;
				break;
			case 'n':
				ignore_nice = 0;
				break;
 			case 'v':
 				verbosity++;
				if (verbosity > 10) verbosity = 10;
 				break;
 			case 'q':
 				verbosity = -1;
 				break;
			case 'c':
				t_per_core = strtol(optarg, NULL, 10);
				if (t_per_core < 1) {
					printf("invalid number of cores/proc");
					help();
					exit(ENOTSUP);
				}
				cores_specified = 1;
				break;
			case 'm':
				pprintf(2,"Ignoring mode parameter.\n");
				break;
			case 's':
				pprintf(2,"Ignoring step parameter.\n");
				break;
			case 'p':
				poll = strtol(optarg, NULL, 10);
				if (poll < 0) {
					printf("poll must be non-negative");
					help();
					exit(ENOTSUP);
				}
				pprintf(2,"Polling every %d msecs\n", poll);
				break;
			case 'u':
				highwater = strtol(optarg, NULL, 10);
				if ((highwater < 0) || (highwater > 100)) {
					printf("upper limit must be between 0 and 100\n");
					help();
					exit(ENOTSUP);
				}
				pprintf(2,"Using upper pct of %d%%\n",highwater);
				break;
			case 'l':
				pprintf(2,"Lower pct ignored.\n");
				break;
			case 'h':
			default:
				help();
				return 0;
		}
	}

	if (daemonize)
		openlog("powerthend", LOG_AUTHPRIV|LOG_PERROR, LOG_DAEMON);

	/* My ego is as big as John's */
	pprintf(0,"PowerThen Daemon v%s, (c) 2005-2006 Peter Schloemer\n",
			VERSION);
	pprintf(0,"Based on PowerNowd (c) 2003-2006 John Clemens");

	/* are we root?? */
	if (getuid() != 0) {
		printf("Duh, you're not root, but you'd have to be to run me.\n");
		exit(EPERM);
	}

	pprintf(1,"Settings:\n");
	pprintf(1,"  verbosity:     %4d\n", verbosity);
	pprintf(1,"  highwater:     %4d %%\n", highwater);
	pprintf(1,"  poll interval: %4d ms\n", poll);

	/*
	 * This should tell us the number of CPUs that Linux thinks we have,
	 * or, at least GLIBC
	 */
	ncpus = sysconf(_SC_NPROCESSORS_CONF);

	if (ncpus < 0) {
		perror("sysconf could not determine number of cpus, assuming 1\n");
		ncpus = 1;
	}

	if (cores_specified) {
		if (ncpus < t_per_core) {
			printf("\nWARNING: bogus # of thread per core, assuming 1\n");
			threads_per_core = 1;
		} else {
			threads_per_core = t_per_core;
		}
	} else {
		threads_per_core = determine_threads_per_core(ncpus);
		if (threads_per_core < 0)
			threads_per_core = 1;
	}

	/* We don't support mixed configs yet */
	if (!ncpus || !threads_per_core || ncpus % threads_per_core) {
		printf("WARN: ncpus(%d) is not a multiple of threads_per_core(%d)!\n",
			ncpus, threads_per_core);
		printf("WARN: Assuming 1.\n");
		threads_per_core = 1;
	}

	num_real_cpus = ncpus/threads_per_core;

	/* Malloc, initialise data structs */
	all_cpus = (cpuinfo_t **) malloc(sizeof(cpuinfo_t *)*ncpus);
	if (all_cpus == (cpuinfo_t **)NULL) {
		perror("Couldn't malloc all_cpus");
		return ENOMEM;
	}

	for (i=0; i<ncpus; i++) {
		all_cpus[i] = (cpuinfo_t *)malloc(sizeof(cpuinfo_t));
		if (all_cpus[i] == (cpuinfo_t *)NULL) {
			perror("Couldn't malloc all_cpus");
			return ENOMEM;
		}
		memset(all_cpus[i],0,sizeof(cpuinfo_t));
	}

	for (i=0;i<ncpus;i++) {
		all_cpus[i]->threads_per_core = threads_per_core;
		all_cpus[i]->scalable_unit = (i/threads_per_core)*threads_per_core;
	}

	pprintf(0,"Found %d scalable unit%s:  -- %d 'CPU%s' per scalable unit\n",
			num_real_cpus,
			(num_real_cpus>1)?"s":"",
			threads_per_core,
			(threads_per_core>1)?"s":"");

	for (i=0;i<ncpus;i++) {
		cpu = all_cpus[i];
		if ((err = get_per_cpu_info(cpu, i)) != 0) {
			printf("\n");
			goto out;
		}
		pprintf(0,"  cpu%d: %dMhz - %dMhz (%d steps)\n",
				cpu->cpuid,
				cpu->min_speed / 1000,
				cpu->max_speed / 1000,
				cpu->table_size);
		for(j=0;j<cpu->table_size; j++) {
			pprintf(1, "     step%d : %ldMhz\n", j+1,
					cpu->freq_table[j] / 1000);
		}
	}

	/* now that everything's all set up, lets set up a exit handler */
	signal(SIGTERM, terminate);
	signal(SIGINT, terminate);

	if (daemonize)
		daemon(0, 0);

	/* Now the main program loop */
	while(1) {
		usleep(poll*1000);
		for(i=0; i<num_real_cpus; i+=j) {
			change = 0;
			cpubase = i*threads_per_core;
			pprintf(6, "i = %d, cpubase = %d, ",i,cpubase);
			/* handle SMT/CMP here */
			for (j=0; j<all_cpus[cpubase]->threads_per_core; j++) {
				change2 = decide_speed(all_cpus[cpubase+j]);
				pprintf(6, "change = %d, change2 = %d\n",change,change2);
				if (change2 > change)
					change = change2;
			}
			change_speed(all_cpus[cpubase], change);
		}
	}

out:
	printf("PowerThend encountered and error and could not start.\n");
	printf("Please make sure that:\n");
	printf(" - You are running a v2.6.7 kernel or later\n");
	printf(" - That you have sysfs mounted /sys\n");
	printf(" - That you have the core cpufreq and cpufreq-userspace\n");
	printf("   modules loaded into your kernel\n");
	printf(" - That you have the cpufreq driver for your cpu loaded,\n");
	printf("   (for example: powernow-k7), and that it works. Check\n");
	printf("   'dmesg' for errors.\n");
	printf("If all of the above are true, and you still have problems,\n");
	printf("please try whether PowerNowd, the program PowerThend is\n");
	printf("based on, works on your system. If it does, please email\n");
	printf("the author: powerthend@scheissname.de. If it doesn't, I\n");
	printf("recommend to try making PowerNowd work first. Most of the\n");
	printf("code comes from there, so the author knows much more about\n");
	printf("what can go wrong, and there is more documentation.\n");

	/* should free more here.. will get to that later.... */
	/* or we can just be lazy and let the OS do it for us... */
	free(cpu);
	return err;
}

